/*
 * IMPORTANT:  READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
 * By downloading, copying, installing or using the software you agree
 * to this license.  If you do not agree to this license, do not download,
 * install, copy or use the software.  
 *
 *
 * Copyright (c) 2006 Northwestern University
 * All rights reserved.  
 *
 * Redistribution of the software in source and binary forms,
 * with or without modification, is permitted provided that the
 * following conditions are met: 
 *
 * 1       Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.  
 *
 * 2       Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.  
 *
 * 3       Neither the name of Northwestern University nor the names of its
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.  
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-INFRINGEMENT AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * NORTHWESTERN UNIVERSITY OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*************************************************************************/
/**   File:         generate.cpp                                        **/
/**   Description:  Describes a data genetaor to generate multi         **/
/**                 dimensional data for clutering. It is based on      **/
/**                 Normal Distribution and the parameters are lower    **/
/**                 and upper bound for the cluster size and standard   **/
/**                 deviation of the cluster radius.                    **/
/**   Author:  Prabhat Kumar                                            **/
/**            Northwestern University.                                 **/
/**            email: pku649@eecs.northwestern.edu                      **/
/**            13 Jan 2010                                              **/
/**                                                                     **/
/*************************************************************************/

#include <time.h>
#include <boost/random/normal_distribution.hpp>
#include <boost/random/mersenne_twister.hpp>
#include <boost/random/variate_generator.hpp>
#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <vector>
#include <math.h>
#include <unistd.h>
#include <float.h>
#include <string.h>

#ifdef _PNETCDF_BUILT
/* for netcdf file */
#include <mpi.h>
#include <pnetcdf.h>
#endif

#include <stdio.h>

#define PI 3.14159

/* Constants for simple cluster centre generation */
#define SIMPLE_RADIUS 20
#define FRAC_UNIFORM 0
#define SIMPLE_STD_DEV_PERCENT 0.08

/* Constants for polar cluster centre generation */
/* Radius of the clusters is within this limits */
#define POLAR_LOWER_THRESH 40
#define POLAR_UPPER_THRESH 80
#define POLAR_STD_DEV_PERCENT 0.15

#define NUM_REC 2000
#define NUM_ATTR 4
#define NUM_CLUS 8
#define POINTS_SPREAD_PERCENT 0.10

using namespace std;
void usage(char *argv0) {
    char const *help =
        "Usage: %s [switches] -o filename -option <default values>\n"
        "       -o filename: output file containing data to be clustered\n"
        "       -n 2000      number of points to be generated\n"
        "       -d 4         number of attributes for each data point\n"
        "       -c 8         number of clusters present in the data set\n"
        "       -b 1         type of file generated. (0 = csv, 1 = binary, 2 = netcdf)\n"
        "       -t 1         type of cluster centre generator used (0=simple, 1=polar)\n"
        "       -l 40        lower threshold for the size of cluster(used only with polar)\n"
        "       -u 80        upper threshold for the size of cluster(used only with polar)\n"
        "       -r 20        radius of cluster(used only with simple)\n"
        "       -s 0.15      std dev percentage of the normal distribution, std dev = s * cluster radius\n"
        "       -f 0         fraction of uniformly distributed points\n"
        "       -v 0.10      variation in the number of points in each cluster, distributed across the centre\n";
        "       -z 0         if 1, random distribution of points will be generated between 0 and 1. default value is 0\n";
    cout << help << endl;
    exit(-1);
}

#ifdef _PNETCDF_BUILT
/* Handle netcdf error */
static void handle_error(int status){
	fprintf(stderr, "%s\n", ncmpi_strerror(status));
}
#endif

/*
 * n = no. of points in the ary, where we have to check the distance of last point to the other remaining points
 * make thres around half of the RANGE, i.e., thresh = 0.5 * RANGE
 * */
bool check_distance(float* ary, int n, int dim, float thresh){

	bool result = true;

	float* point = (float*)malloc(sizeof(float) * dim);

	/* Make a copy of the last point */
	for(int i = 0 ; i < dim; i++){
		point[i] = ary[(n-1) * dim + i];
	}

	float dist;
	/* Check the distance from all the points */
	for(int i = 0; i < (n-1); i++){
		dist = 0.0;
		for(int j = 0; j < dim; j++){
			dist += (point[j] - ary[i * dim + j]) * (point[j] - ary[i * dim + j]);
		}
		dist = sqrt(dist);
		//cout << "DIST: " << dist << endl;

		if(dist < thresh){
			result = false;
			break;
		}

	}

	/* Return */
	return result;
}

/* Compute ditance between two points */
// This one is currently used
bool check_distance_1(float* ref, float* point, int dim, float thresh){

	bool result = true;

	float dist;
	for(int j = 0; j < dim; j++){
		dist += (ref[j] - point[j]) * (ref[j] - point[j]);
	}
	dist = sqrt(dist);

	if(dist < thresh) result = false;

	/* Return */
	return result;
}

template<class T>
T gen_normal_4(T &generator, int total_dim, float* ary, int dim, int start, int end){

	for(int i = start; i < end; i++){
		//cout << "TEST : " << i << endl;
		ary[i * total_dim + dim] = generator();
	}

	return generator;
}

void print_ary(float* ary, int row, int col){
	for(int i = 0; i < row; i++){
		for(int j = 0; j < col; j++){
			cout <<ary[i * col + j] << ",";
		}
		cout << endl;
	}
}

void print_ary_int(int* ary, int row, int col){
	for(int i = 0; i < row; i++){
		for(int j = 0; j < col; j++){
			cout <<","<<ary[i * col + j] << ",";
		}
		cout << endl;
	}
}

/* Function to compute cluster centres 
 * Pick random points from uniformaly generated points and then compute the distance between them which should be greater than THRESH
 * */
void generate_cluster_centres_simple(float* centres, float* cluster_radius, float* records, int uniform_points, int num_clusters, int num_dims){
	
	int index;
	
	/* Initialize cluster radius */
	for(int i = 0; i < num_clusters; i++){
		cluster_radius[i] = SIMPLE_RADIUS;
	}

	for(int i = 0; i < num_clusters; i++){

		index = rand() % uniform_points;

		for(int dim = 0; dim < num_dims; dim++){
			centres[i * num_dims + dim] = records[index * num_dims + dim];
		}

		/* i = number of cluster generated till now */
		for(int j = 0; j < i; j++){
			if (!check_distance_1(&centres[i * num_dims], &centres[j * num_dims], num_dims, cluster_radius[i] + cluster_radius[j])){
				i--;
				break;
			}
		}
	}
}

/* New function to generate cluster centres based on polar coordinates 
 * Origin (0, 0, 0, .., 0) is chosen as the first cluster and then based on the size of the cluster ( determined by POLAR_LOWER_THRESH and POLAR_UPPER_THRESH)
 * choose other cluster centres. Once k clusters is chosen the reference point is one of the k centres to choose k+1 centre
 * */
void generate_cluster_centres_polar(float* centres, float* cluster_radius, int num_clusters, int num_dims, float polar_lower_thresh, float polar_upper_thresh){

	int ref_cluster = 0;
	float theta;

	/* Initialize cluster radius */
	for(int i = 0; i < num_clusters; i++){
		cluster_radius[i] = polar_lower_thresh + (polar_upper_thresh - polar_lower_thresh) * ((double)rand()/RAND_MAX);
	}
	//print_ary(cluster_radius, num_clusters, 1);

	/* Choose one centre to be the origin, We will decide later how to generalize this */
	for(int dim = 0; dim < num_dims; dim++) centres[dim] = 0;

	for(int i = 1; i < num_clusters; i++){
		/* Pick one of the cluster centres which are already generated and use this to generate another one */
		ref_cluster = rand() % i;

		/* First two dimensions */
		theta = 2 * PI * ((double)rand()/RAND_MAX); // Just to make sure that cos and sin covers all the range
		centres[i * num_dims] = centres[ref_cluster * num_dims] + (cluster_radius[ref_cluster] + cluster_radius[i]) * cosf(theta);
		centres[i * num_dims + 1] = centres[ref_cluster * num_dims + 1] + (cluster_radius[ref_cluster] + cluster_radius[i]) * sinf(theta);
		/* Remaining dimensions */
		for(int dim = 2; dim < num_dims; dim++){
			theta = PI * ((double)rand()/RAND_MAX);
			centres[i * num_dims + dim] = centres[ref_cluster * num_dims + dim] + (cluster_radius[ref_cluster] + cluster_radius[i]) * cosf(theta);
		}
			
		if(!check_distance_1(&centres[i * num_dims], &centres[ref_cluster * num_dims], num_dims, cluster_radius[ref_cluster] + cluster_radius[i])){
			i--;
			break;
		}
	}

}

/* Compute the minimum and maximum in all the dimensions for all the clusters */
void compute_dim_range(float* dim_range, float* centres, int num_dims, int num_clusters){
	
	/* Compute maximum and minimum of all the dimensions */
	for(int i = 0; i < num_dims; i++){ 
		dim_range[2 * i] = FLT_MAX;
		dim_range[2 * i + 1] = FLT_MIN;
	}
	for(int i = 0; i < num_dims; i++){
		for(int j = 0; j < num_clusters; j++){
			if(dim_range[2 * i] > centres[j * num_dims + i]) dim_range[2 * i] = centres[j * num_dims + i];
			if(dim_range[2 * i + 1] < centres[j * num_dims + i]) dim_range[2 * i + 1] = centres[j * num_dims + i];
		}
	}
}

/* Main program */
int main(int argc, char** argv){

	int isBinary = 1;
	int cluster_gen_type = 1; // 0 for simple and 1 for polar
	int num_points = NUM_REC;
	int num_dims = NUM_ATTR;
	int num_clusters = NUM_CLUS;
	float polar_lower_thresh = POLAR_LOWER_THRESH;
	float polar_upper_thresh = POLAR_UPPER_THRESH;
	float simple_radius = SIMPLE_RADIUS;
	float frac_uniform = FRAC_UNIFORM;
	float simple_std_dev_percent = SIMPLE_STD_DEV_PERCENT;
	float polar_std_dev_percent = POLAR_STD_DEV_PERCENT;
	float points_spread_percent = POINTS_SPREAD_PERCENT;
	char* filename;
	char  stats_filename[1024];
	int isRandom = 0;
	
	int opt;
	extern char* optarg;
	extern int optind, optopt;
	while( (opt = getopt(argc, argv, "o:n:d:c:b:t:l:u:r:s:f:v:z:")) != -1 ){
		switch(opt){
			case 'o': filename = optarg;
					  break;
			case 'n': num_points = atoi(optarg); // Number of points
					  break;
			case 'd': num_dims = atoi(optarg); // Number of attributes
					  break;
			case 'c': num_clusters = atoi(optarg); // Number of clusters
					  break;
			case 'b': isBinary = atoi(optarg); // is the file binary
					  break;
			case 't': cluster_gen_type = atoi(optarg); // what type of cluster centre genetor u are using
					  break;
			case 'l': polar_lower_thresh = atof(optarg); // lower threashold for polar generator
					  break;
			case 'u': polar_upper_thresh = atof(optarg); // upper threshold for polar generator
					  break;
			case 'r': simple_radius = atof(optarg); // radius for simple generator
					  break;
			case 's': polar_std_dev_percent = simple_std_dev_percent = atof(optarg); // std dev
					  break;
			case 'f': frac_uniform = atof(optarg); // if you want uniform points		  
					  break;
			case 'v': points_spread_percent = atof(optarg); // variation in the number of points in different cluster
					  break;
			case 'z': isRandom = atoi(optarg); // If just a random ditrbituion of points is required
					  break;
			case '?': usage(argv[0]);		  
					  break;
			default:  usage(argv[0]);	
					  break;
		}
	}

	/* for netcdf file output 
	 * */
	int status;
	int ncid;
	int records_dim[2];
	int records_id;
	static char title[] = "netCDF dataset";
	static char description[] = "2-D array containing records and attributes";
	int rank, nprocs;

#ifdef _PNETCDF_BUILT
	/* MPI Stuff used for pnetcdf, though I am using only one node */
	MPI_Init(&argc, &argv);
	MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#endif

	if(filename == 0) usage(argv[0]);

	float* records = (float*)malloc(sizeof(float) * num_points * num_dims);
	float* centres = (float*)malloc(sizeof(float) * num_clusters * num_dims);
	int* histogram = (int*)malloc(sizeof(int) * (num_clusters + 1));
	float* cluster_radius = (float*)malloc(sizeof(float) * num_clusters);
	float* dim_range = (float*)malloc(sizeof(float) * num_dims * 2); // for max and min in each dimension


	cout << "POINTS: [" << num_points << "] DIMS: [" << num_dims << "] CLUSTERS: [" << num_clusters << "]"<< endl;

	/* Generate 10 times the number of clusters
	 * and the range would be clusters * radius and half of this humber
	 * */
	//srand(time(NULL));
	srand(123456);
	int sample_cluster_points = 50 * num_clusters;
	for(int i = 0; i < sample_cluster_points; i++){
		for(int dim = 0; dim < num_dims; dim++){
			records[i * num_dims + dim] = (0.5 * num_clusters * simple_radius) * (double)rand()/RAND_MAX;
		}
	}
	//print_ary(records, 10, num_dims);

	/* Pick num_clusters points from the unformly generated random points: Put a distance criterion */
	int index;
	float cur_std_dev;

	cout << "Generating cluster centres... "<< endl;
	if(cluster_gen_type == 0){
		generate_cluster_centres_simple(centres, cluster_radius, records, sample_cluster_points, num_clusters, num_dims);
		cur_std_dev = simple_std_dev_percent;
	} else if(cluster_gen_type == 1){
		generate_cluster_centres_polar(centres, cluster_radius, num_clusters, num_dims, polar_lower_thresh, polar_upper_thresh);
		cur_std_dev = polar_std_dev_percent;
	} else {
		cout << "Unrecognized Cluster Generator Type. Exiting..." << endl;
		exit(0);
	}
	cout << "Cluster Centres generated."<< endl;

	int uniform_points = (int)(frac_uniform * num_points);
	if(uniform_points > 0){
		/* Compute the range of the minimum and the maximum distance */
		compute_dim_range(dim_range, centres, num_dims, num_clusters);
		
		/* Generate random distribution of points depending on the percentage of points to be uniformly generated */
		for(int i = 0; i < uniform_points; i++){
			for(int dim = 0; dim < num_dims; dim++){
				records[i * num_dims + dim] = dim_range[2*dim] + (dim_range[2*dim + 1] - dim_range[2*dim]) * (double)rand()/RAND_MAX; // dim_range[0] is the lower and dim_range[1] is the upper and so forth
			}
		}
		//print_ary(dim_range, num_dims, 2);
	}


	/* The start and end of the different cluster centres */
	int points = (num_points - uniform_points)/num_clusters;
	histogram[0] = uniform_points;
	//int points = num_points/num_clusters;
	//histogram[0] = 0;
	float frac_variation;
	for(int i = 1; i < num_clusters; i++){
		frac_variation = (-1 * points_spread_percent) + 2 * points_spread_percent * ((double)rand()/RAND_MAX);
		histogram[i] = points + (int)(points * frac_variation) + histogram[i-1];
	}
	histogram[num_clusters] = num_points;
	//print_ary_int(histogram, num_clusters + 1, 1);
	//exit(0);

	/* Generate normaly distribution points */
	cout << "Generating Data Points ... "<< endl;
	if(isRandom == 1){
		for(int p = 0; p < num_points; p++){
			for(int dim = 0; dim < num_dims; dim++){
				records[p * num_dims + dim] = (double)rand()/(double)RAND_MAX;
			}
		}
	} else if(isRandom == 0) {
		for(int dim = 0; dim < num_dims; dim++){
			for(int j = 0; j < num_clusters; j++){
				boost::variate_generator<boost::mt19937, boost::normal_distribution<> >
					generator(boost::mt19937(100 + (dim+1)*(j+1)), boost::normal_distribution<>(centres[j * num_dims + dim], cur_std_dev * cluster_radius[j]));
				//boost::variate_generator<boost::mt19937, boost::normal_distribution<> >
				//	generator(boost::mt19937(100 + (dim+1)*(j+1)), boost::normal_distribution<>(centres[j * num_dims + dim], STD_DEV));
				
				generator = gen_normal_4(generator, num_dims, records, dim, histogram[j], histogram[j+1]);
			}	
		}
	}
	//for(int dim = 0; dim < num_dims; dim++){
	//	for(int j = 0; j < num_clusters; j++){
	//		boost::variate_generator<boost::mt19937, boost::normal_distribution<> >
	//			generator(boost::mt19937(100 + (dim+1)*(j+1)), boost::normal_distribution<>(centres[j * num_dims + dim], cur_std_dev * cluster_radius[j]));
	//		//boost::variate_generator<boost::mt19937, boost::normal_distribution<> >
	//		//	generator(boost::mt19937(100 + (dim+1)*(j+1)), boost::normal_distribution<>(centres[j * num_dims + dim], STD_DEV));
	//		
	//		generator = gen_normal_4(generator, num_dims, records, dim, histogram[j], histogram[j+1]);
	//	}	
	//}
	cout << "Data points generated. "<< endl;
	//print_ary(records, num_points, num_dims);

	/* Randominze the data for writing to a file */
	int* random_order = (int*)malloc(sizeof(int) * num_points);
	int* uniform_order = (int*)malloc(sizeof(int) * num_points);
	

	int count = 0;
	for(int i = 0; i < num_points; i++){
		random_order[i] = -1;
		uniform_order[i] = i;
	}
	
	for(int i = 0; i < num_points; i++){
		index = rand() % num_points;
		
		if(uniform_order[index] != -1){
			random_order[i] = index;
			uniform_order[index] = -1;
		} else {
			i--;
		}
	}
	//print_ary_int(random_order, num_points, 1);

	ofstream outFile;
	FILE *fp;

	if(isBinary == 1){
		cout << "Writing data to a Binary file: " << filename <<endl;
		fp = fopen(filename, "wb");

		fwrite(&num_points, sizeof(int), 1, fp);
		fwrite(&num_dims, sizeof(int), 1, fp);
		fwrite(records, sizeof(float), num_points * num_dims, fp);

		/* THIS IS A BUG HERE.. The data written in this way is not read exactly as that way */
		/* Write number of data points and then the number of dimensions int he binary file */
		outFile.write((char *)&num_points, sizeof(int));
		outFile.write((char *)&num_dims, sizeof(int));

		/* Now write the data */
		for(int i = 0; i < num_points; i++){
			index = random_order[i];
			for(int j = 0; j < (num_dims-1); j++){
				//cout << "Write Data: " << records[index * num_dims + j] << " ";
				outFile.write((char *)&records[index * num_dims + j], sizeof(float));
			}
			outFile.write((char*)&records[index * num_dims + (num_dims - 1)], sizeof(float));

			//cout << "Write Data: " << records[index * num_dims + (num_dims - 1)] << " " << endl;
		}
		/* close the file */
		cout << "Data file written."<< endl;
		fclose(fp);

	}
#ifdef _PNETCDF_BUILT
	else if(isBinary == 2){
		MPI_Offset records_start[2], records_count[2];
		cout << "Writing data to a netcdf file: " << filename <<endl;

		status = ncmpi_create(MPI_COMM_WORLD, filename, NC_CLOBBER, MPI_INFO_NULL, &ncid);
		if(status != NC_NOERR) handle_error(status);

		status = ncmpi_put_att_text(ncid, NC_GLOBAL, "title", strlen(title), title);
		if(status != NC_NOERR) handle_error(status);

		status = ncmpi_def_dim(ncid, "Records", num_points, &records_dim[0]);
		if(status != NC_NOERR) handle_error(status);
		status = ncmpi_def_dim(ncid, "Attributes", num_dims, &records_dim[1]);
		if(status != NC_NOERR) handle_error(status);

		status = ncmpi_def_var(ncid, "objects", NC_FLOAT, 2, records_dim, &records_id);
		if(status != NC_NOERR) handle_error(status);

		status = ncmpi_put_att_text(ncid, records_id, "description", strlen(description), description);
		if(status != NC_NOERR) handle_error(status);

		status = ncmpi_enddef(ncid);
		if(status != NC_NOERR) handle_error(status);

		records_start[0] = rank * (num_points/nprocs);
		records_start[1] = 0;
		records_count[0] = num_points/nprocs;
		records_count[1] = num_dims;

		status = ncmpi_put_vara_float_all(ncid, records_id, records_start, records_count, &records[rank * records_count[0] * records_count[1]]);
		if(status != NC_NOERR) handle_error(status);

		cout << "Data file written."<< endl;
		status = ncmpi_close(ncid);
		if(status != NC_NOERR) handle_error(status);
	}
#endif
	else {
		cout << "Writing data to a CSV(comma seperated file) file: " << filename <<endl;
		//outFile.open("./data/data.csv", ios::out);
		outFile.open(filename, ios::out);

		/* Now write the data */
		for(int i = 0; i < num_points; i++){
			index = random_order[i];
			for(int j = 0; j < (num_dims-1); j++){
				outFile << records[index * num_dims + j] << ",";
			}
			outFile << records[index * num_dims + (num_dims-1)] << endl;
		}
		/* close the file */
		cout << "Data file written."<< endl;
		outFile.close();
	}

	///* Print a couple of data elements */
	//for(int i = 0; i < 10; i++){
	//	for(int j = 0; j < num_dims; j++){
	//		cout << records[i * num_dims + j];
	//	}
	//	cout << endl;
	//}

	/* Testing.. printing the data */
	//print_ary(records, num_points, num_dims);
	
	/* Print the relevant stats */
        sprintf(stats_filename, "%s.stats",filename);
	cout << "Writing data stats to the file: " << stats_filename << endl;
	ofstream statsFile;
	statsFile.open(stats_filename, ios::out);
	statsFile << "Num of data points: " << num_points << endl;
	statsFile << "Num of attributes : " << num_dims << endl;
	statsFile << "Number of cluster : " << num_clusters << endl;
	statsFile << "Number of Uniformly generated points : " << uniform_points << endl;
	statsFile << "Cluster Centres Info : " << endl;
	for(int i = 0; i < num_clusters; i++){
		statsFile << "Cluster[" << i << "]" << " [" << histogram[i+1] - histogram[i] << "] ";
		for(int j = 0; j < num_dims; j++){
			statsFile << centres[i * num_dims + j] << " ";
		}
		statsFile << endl;
	}
	statsFile.close();

#ifdef _PNETCDF_BUILT
	MPI_Finalize();
#endif
	cout << "Exiting..." << endl;
}
